@@ -25,6 +25,14 @@ class Awk_command:
         super(Awk_command, self).__setattr__('persistance_in_time', "continuous")
         super(Awk_command, self).__setattr__('persistance_after_execution', "ephemeral")
         super(Awk_command, self).__setattr__('id', get_random_string(20))
+        super(Awk_command, self).__setattr__('output_header', -1)
+        super(Awk_command, self).__setattr__('has_header', -1)
+        super(Awk_command, self).__setattr__('nrow_selected', 0)
+        super(Awk_command, self).__setattr__('ncol_selected', 0)
+        super(Awk_command, self).__setattr__('nrow_before', 0)
+        super(Awk_command, self).__setattr__('ncol_before', 0)
+        
+
     def __getattr__(self, name):
         return super(Awk_command, self).__getattr__(name)
     def __setattr__(self, name, value):
@@ -33,25 +41,20 @@ class Awk_command:
 
 class DataFrame:
     def __repr__(self):
-        text = self.__head_current__().values()
-        shape = self.__shape_current__().values()
-        values = shape.replace("\n","").strip().split(",")
-        values = np.array(values).astype(int)
-        if values[0] > 10:
+        text = self.__head_current__().values()        
+        if self.nrow > 10:
             text = text + ("...")
         return text
     def __str__(self):
-        text = self.__head_current__().values()
-        shape = self.__shape_current__().values()
-        values = shape.replace("\n","").strip().split(",")
-        values = np.array(values).astype(int)
-        if values[0] > 10:
+        text = self.__head_current__().values()        
+        if self.nrow > 10:
             text = text + ("...")
         return text
     def __new__(cls, *args, **kwargs):
         return super().__new__(cls)
     def __init__(self):
         super(DataFrame, self).__setattr__('path', "")
+        super(DataFrame, self).__setattr__('has_header',True)
         super(DataFrame, self).__setattr__('commands', [])
         super(DataFrame, self).__setattr__('delimiter', ",")
         super(DataFrame, self).__setattr__('id', get_random_string(20))
@@ -59,6 +62,9 @@ class DataFrame:
         super(DataFrame, self).__setattr__('nrow', 0)
         super(DataFrame, self).__setattr__('__ncol_original__', 0)
         super(DataFrame, self).__setattr__('__nrow_original__', 0)
+        super(DataFrame, self).__setattr__('columns', [])
+        super(DataFrame, self).__setattr__('selected_columns', [])
+        super(DataFrame, self).__setattr__('nrow_modified_to_unknown_value', True)
 
     def __getattr__(self, name):
         return super(DataFrame, self).__getattr__(name)
@@ -119,47 +125,62 @@ class DataFrame:
 #     def __setitem__(self,key,values):
 #         DataFrame.__setattr__(self,key,values)
 
-    def read_csv(self,path,delimiter = ","):
+    def read_csv(self,path,delimiter = ",",has_header = True,names_columns = []):
         self.path = path
         self.delimiter = delimiter
+        self.has_header = has_header
+        if names_columns != []:
+            self.columns = names_columns
+        else:
+            self.names()
+        self.shape()
 
     def shape(self):
-        input = self.path
-        command = "awk " + """'
-        BEGIN {
-            FS = \"""" + self.delimiter + """\"
-            number_rows = 0
-        }
-        {
-            if (NR == 1){
+        if self.__ncol_original__ == 0 and self.__nrow_original__ == 0:
+            command = "wc -l " + self.path
+            result = os.popen(command).read()
+            values = result.replace("\n","").strip().split(" ")[0]
+            nrow = int(values)
+            if self.has_header:
+                nrow -= 1
+                
+            command = "awk 'BEGIN{FS=\"" + self.delimiter + "\"}{if (NR>1){exit}}END{print NF}' " + self.path
+            result = os.popen(command).read()
+            values = result.replace("\n","").strip()
+            ncol = int(values)
+            values = np.array([nrow,ncol]).astype(int)
+            self.ncol = values[1]
+            self.nrow = values[0]
+            self.__ncol_original__ = ncol
+            self.__nrow_original__ = nrow
+            return values
+        else:
+            if self.nrow_modified_to_unknown_value:
+                shape = self.__shape_current__().values()
+                values = shape.replace("\n","").strip().split(",")
+                values = np.array(values).astype(int)
+#                 self.nrow_modified_to_unknown_value = False
+                self.ncol = values[1]
+                self.nrow = values[0]
+                self.__ncol_original__ = self.ncol
+                self.__nrow_original__ = self.nrow
+                return values
+            else:
+                return [nrow,ncol]
 
-            }else{
-                number_rows+=1
 
-            }
-        }
-        END {
-            printf("%s""" + self.delimiter + """",number_rows)
-            print(NF)
-        }
-        ' """ + input
-        result = os.popen(command).read()
-        values = result.replace("\n","").strip().split(",")
-        values = np.array(values).astype(int)
-        ncol = values[1]
-        nrow = values[0]
-        __ncol_original__ = ncol
-        __nrow_original__ = nrow
-        return values
-
-    def __shape_current__(self):
-        command = "awk " + """'
+    def __shape_current__(self,has_header = True):
+        command = "awk -v has_header=has_header_variable " + """'
         BEGIN {
             FS = \"""" + self.delimiter + """\"
             number_rows = 0
         }
         {
             if (NR == 1){
+                if (has_header){
+                }else{
+                    number_rows+=1
+                }
 
             }else{
                 number_rows+=1
@@ -171,7 +192,6 @@ class DataFrame:
             print(NF)
         }
         ' """
-        result = os.popen(command).read()
 
         awk_command = Awk_command()
         awk_command.command = command
@@ -184,44 +204,51 @@ class DataFrame:
 
 
     def names(self):
-        input = self.path
-        command = "awk " + """'
-    BEGIN {
-        FS = \"""" + self.delimiter + """\"
-    }
-    {
-        if (NR == 1){
-            print
-            exit
-        }
-    }
-    END {}
-    ' """ + input
-        result = os.popen(command).read()
-        return np.array(result.replace("\n","").split(self.delimiter))
+        if len(self.columns) == 0:
+            if self.has_header:
+                input = self.path
+                command = "awk " + """'
+            BEGIN {
+                FS = \"""" + self.delimiter + """\"
+            }
+            {
+                if (NR == 1){
+                    print
+                    exit
+                }
+            }
+            END {}
+            ' """ + input
+                result = os.popen(command).read()
+                self.columns = np.array(result.replace("\n","").split(self.delimiter))
+                return self.columns
+            else:
+                return []
+        else:
+            return self.columns
 
-    def __names_current__(self):
-        command = "awk " + """'
-    BEGIN {
-        FS = \"""" + self.delimiter + """\"
-    }
-    {
-        if (NR == 1){
-            print
-            exit
-        }
-    }
-    END {}
-    ' """
-        awk_command = Awk_command()
-        awk_command.command = command
-        awk_command.priority = 9999999999
-        awk_command.type = "names"
-        awk_command.persistance_in_time = "instance"
-        self.commands.append(awk_command)
-        return self
+#     def __names_current__(self):
+#         command = "awk " + """'
+#     BEGIN {
+#         FS = \"""" + self.delimiter + """\"
+#     }
+#     {
+#         if (NR == 1){
+#             print
+#             exit
+#         }
+#     }
+#     END {}
+#     ' """
+#         awk_command = Awk_command()
+#         awk_command.command = command
+#         awk_command.priority = 9999999999
+#         awk_command.type = "names"
+#         awk_command.persistance_in_time = "instance"
+#         self.commands.append(awk_command)
+#         return self
 
-    def __to_np_arrays__(self):
+    def __to_np_arrays__(self,has_header = True):
         shape = self.shape()
         columns = np.empty(shape[0],list)
         lines = self.values(clear = False).split("\n")
@@ -231,12 +258,20 @@ class DataFrame:
         for type_t  in types_text:
             types.append(eval(type_t))
         line_counter = 0
-        names = []
+        names = self.columns
         for line in lines:
             if line != "":
                 elements = line.split(self.delimiter)
                 if line_counter == 0:
-                        names = elements
+                        if has_header:
+                            names = elements
+                        else:
+                            for i in range(len(elements)):
+                                try:
+                                    columns[i].append(elements[i])
+                                except:
+                                    columns[i] = [elements[i]]
+                            
                 else:
                     for i in range(len(elements)):
                         try:
@@ -262,11 +297,25 @@ class DataFrame:
         return df
 
     def values(self,clear = True):
+        if self.has_header:
+            has_header = 1
+        else:
+            has_header = 0
+        output_header = has_header
         complete_command = ""
         record_delimiter = "\\n"
         record_delimiter_transform = "\\n"
         intermediate_record_delimiter = "\\n"
         for command in self.commands:
+            if output_header == 0:
+                has_header = 0
+            if command.has_header != -1:
+                if command.has_header == 0:
+                    has_header = 0
+            if command.output_header != -1:
+                if command.output_header == 0:
+                    output_header = 0
+            
             if command == self.commands[0]:
                 if command != self.commands[len(self.commands)-1]:
                     record_delimiter_transform = "\\n"
@@ -274,14 +323,17 @@ class DataFrame:
                     record_delimiter_transform = intermediate_record_delimiter
                 if command == self.commands[len(self.commands)-1]:
                     record_delimiter_transform = "\\n"
-                complete_command = command.command.replace("has_header","1").replace("record_delimiter_transform",record_delimiter_transform).replace("record_delimiter",record_delimiter) + self.path
+                
+                
+                       
+                complete_command = command.command.replace("has_header_variable",str(has_header)).replace("output_header_variable",str(output_header)).replace("record_delimiter_transform",record_delimiter_transform).replace("record_delimiter",record_delimiter) + self.path
                 record_delimiter = record_delimiter_transform
             else:
                 if command != self.commands[len(self.commands)-1]:
                     record_delimiter_transform = intermediate_record_delimiter
                 else:
                     record_delimiter_transform = "\\n"
-                complete_command = complete_command + " | " + command.command.replace("has_header","1").replace("record_delimiter_transform",record_delimiter_transform).replace("record_delimiter",record_delimiter)
+                complete_command = complete_command + " | " + command.command.replace("has_header_variable",str(has_header)).replace("output_header_variable",str(output_header)).replace("record_delimiter_transform",record_delimiter_transform).replace("record_delimiter",record_delimiter)
                 record_delimiter = record_delimiter_transform
         result = os.popen(complete_command).read()
         new_commands = []
@@ -293,98 +345,62 @@ class DataFrame:
             self.clear_commands()
         return result
 
-    def __get_rows_from_to__(self,min_row,max_row,header = True):
+    def __get_rows_from_to__(self,min_row,max_row,has_header = True,output_header = True):
+
         awk_command = Awk_command()
-#         if type(rows) == list:
-#             rows = np.array(rows)
-#         rows_str = np.array2string(rows,separator="\n")
-#         rows_str = rows_str[1:len(rows_str)-1]
-#         path_rows = os.path.expanduser('~') + "/.tmp/rows_" + self.id + ".txt"
-#         command = "echo '" + rows_str + "'>" + path_rows
-#         os.system(command)
-        variables = "-v header=\"has_header\" -v min_row=\"" + str(min_row) + "\" -v max_row=\"" + str(max_row) + "\" "
-        command = "awk " + variables + """'
-        BEGIN {
-            FS = \"""" + self.delimiter + """\"
-            RS = "record_delimiter"
-            RS_new = "record_delimiter_transform"
+        
+        if not self.nrow_modified_to_unknown_value:
+            if min_row < 0:
+                min_row = 0
+            if max_row > nrow:
+                max_row = nrow        
+            awk_command.nrow_selected = max_row-min_col
+            awk_command.nrow_before = nrow
+            awk_command.ncol_before = ncol
+            self.nrow = awk_command.nrow_selected
 
-        }
-        {
-            if (FNR == 1){
-                if (header){
-                    printf("%s" RS_new,$0)
-                }
-            }else{
-                if (FNR >= min_row + 1 && FNR <= max_row + 1){
-                    printf("%s" RS_new,$0)
-                }
-            }
+        if self.nrow > 0:
+    #         if type(rows) == list:
+    #             rows = np.array(rows)
+    #         rows_str = np.array2string(rows,separator="\n")
+    #         rows_str = rows_str[1:len(rows_str)-1]
+    #         path_rows = os.path.expanduser('~') + "/.tmp/rows_" + self.id + ".txt"
+    #         command = "echo '" + rows_str + "'>" + path_rows
+    #         os.system(command)
 
-            if (FNR > max_row + 1){
-                exit
-            }
-
-        }
-        END {
-
-        }
-        ' """
-        awk_command.command = command
-        awk_command.priority = 1
-        awk_command.type = "get_rows_range"
-        self.commands.append(awk_command)
-        __ddf__ = deepcopy(self)
-        __ddf__.__settle_commands__()
-        self.clear_commands()
-        return __ddf__
-
-    def get_rows(self,rows,header = True):
-        if type(rows) == range:
-            return DataFrame.__get_rows_from_to__(self,min(rows),max(rows))
-        else:
-            awk_command = Awk_command()
-            if type(rows) == list:
-                rows = np.array(rows)
-
-            rows_str = np.array2string(rows,separator="\n")
-            rows_str = rows_str[1:len(rows_str)-1]
-            if not os.path.exists(os.path.expanduser('~') + "/.tmp/"):
-                os.mkdir(os.path.expanduser('~') + "/.tmp/")
-                print("Creating folder ",os.path.expanduser('~') + "/.tmp/")
-            path_rows = os.path.expanduser('~') + "/.tmp/rows_" + self.id + "_" + awk_command.id + ".txt"            
-            command = "echo '" + rows_str + "'>" + path_rows
-            os.system(command)
-            variables = "-v header=\"has_header\" -v min_row=\"" + str(min(rows)) + "\" -v max_row=\"" + str(max(rows)) + "\" -v length_rows=\"" + str(len(rows)) + "\" "
+            variables = "-v has_header=\"has_header_variable\" -v output_header=\"output_header_variable\" -v min_row=\"" + str(min_row) + "\" -v max_row=\"" + str(max_row) + "\" "
             command = "awk " + variables + """'
             BEGIN {
                 FS = \"""" + self.delimiter + """\"
                 RS = "record_delimiter"
                 RS_new = "record_delimiter_transform"
-                rows["1"] = 1
-                cmd = "cat """ + path_rows + """\"
 
-                while (cmd | getline) {
-                    rows[$0+1] = 1;
 
-                }
-
-                close(cmd)
 
             }
             {
-
-                if (rows[FNR] == 1){
+                if (has_header){
                     if (FNR == 1){
-                        if (header){
+                        if (output_header){
                             printf("%s" RS_new,$0)
                         }
                     }else{
-                        printf("%s" RS_new,$0)
+                        if (FNR >= min_row + 2 && FNR <= max_row + 2){
+                            printf("%s" RS_new,$0)
+                        }
+                    }
+
+                    if (FNR > max_row + 2){
+                        exit
+                    }
+                }else{
+                    if (FNR >= min_row + 1 && FNR <= max_row + 1){
+                            printf("%s" RS_new,$0)
+                    }
+
+                    if (FNR > max_row + 1){
+                        exit
                     }
-                }
-                if (FNR > max_row + 1){
-                    exit
                 }
 
             }
@@ -393,6 +409,113 @@ class DataFrame:
             }
             ' """
             awk_command.command = command
+            if output_header:
+                awk_command.output_header = 1
+            else:
+                awk_command.output_header = 0
+            if has_header:
+                awk_command.has_header = 1
+            else:
+                awk_command.has_header = 0
+            awk_command.priority = 1
+            awk_command.type = "get_rows_range"
+            self.commands.append(awk_command)
+            __ddf__ = deepcopy(self)
+            __ddf__.__settle_commands__()
+            self.clear_commands()
+            return __ddf__
+        else:
+            raise Exception("No rows selected")
+
+    def get_rows(self,rows,has_header = True,output_header = True):
+        
+        
+        
+        if type(rows) == range:
+            return DataFrame.__get_rows_from_to__(self,min(rows),max(rows),has_header = has_header,output_header = output_header)
+        else:
+            
+            awk_command = Awk_command()
+            if type(rows) == list:
+                rows = np.array(rows)
+            
+            if not self.nrow_modified_to_unknown_value:
+                rows = rows[np.where(np.isin(rows,np.arange(0,nrow)))[0]]
+                awk_command = Awk_command()
+                awk_command.nrow_selected = len(rows)
+                awk_command.nrow_before = nrow
+                awk_command.ncol_before = ncol
+                self.nrow = awk_command.nrow_selected
+            
+            rows_str = np.array2string(rows,separator="\n")
+            rows_str = rows_str[1:len(rows_str)-1]
+            if not os.path.exists(os.path.expanduser('~') + "/.tmp/"):
+                os.mkdir(os.path.expanduser('~') + "/.tmp/")
+                print("Creating folder ",os.path.expanduser('~') + "/.tmp/")
+            path_rows = os.path.expanduser('~') + "/.tmp/rows_" + self.id + "_" + awk_command.id + ".txt"            
+            command = "echo '" + rows_str + "'>" + path_rows
+            os.system(command)
+            variables = "-v has_header=\"has_header_variable\" -v output_header=\"output_header_variable\" -v min_row=\"" + str(min(rows)) + "\" -v max_row=\"" + str(max(rows)) + "\" -v length_rows=\"" + str(len(rows)) + "\" "
+            command = "awk " + variables + """'
+            BEGIN {
+                FS = \"""" + self.delimiter + """\"
+                RS = "record_delimiter"
+                RS_new = "record_delimiter_transform"
+                
+                cmd = "cat """ + path_rows + """\"
+
+                while (cmd | getline) {
+                    if (has_header){
+                        rows["1"] = 1
+                        rows[$0+2] = 1;
+                    }else{
+                        rows[$0+1] = 1;
+                    }
+
+                }
+
+                close(cmd)
+
+            }
+            {
+                if (has_header){
+                    if (rows[FNR] == 1){
+                        if (FNR == 1){
+                            if (output_header){
+                                printf("%s" RS_new,$0)
+                            }
+                        }else{
+                            printf("%s" RS_new,$0)
+                        }
+                    }
+                    if (FNR > max_row + 2){
+                        exit
+                    }
+                }else{
+                    if (rows[FNR] == 1){
+                        
+                        printf("%s" RS_new,$0)
+                                            }
+                    if (FNR > max_row + 1){
+                        exit
+                    }
+                
+                }
+
+            }
+            END {
+
+            }
+            ' """
+            awk_command.command = command
+            if output_header:
+                awk_command.output_header = 1
+            else:
+                awk_command.output_header = 0
+            if has_header:
+                awk_command.has_header = 1
+            else:
+                awk_command.has_header = 0
             awk_command.priority = 1
             awk_command.type = "get_rows"
             self.commands.append(awk_command)
@@ -401,7 +524,7 @@ class DataFrame:
             self.clear_commands()
             return __ddf__
 
-    def get_cols(self,cols,header = True):
+    def get_cols(self,cols,output_has_header = True,output_header = True):
         awk_command = Awk_command()
         if type(cols) == np.array:
             cols = cols.tolist()
@@ -417,6 +540,13 @@ class DataFrame:
             else:
                 new_cols[i] = col
         cols = np.unique(new_cols)
+        
+        cols = cols[np.where(np.isin(cols,np.arange(0,ncol)))[0]]
+        awk_command = Awk_command()
+        awk_command.ncol_selected = len(cols)
+        awk_command.nrow_before = nrow
+        awk_command.ncol_before = ncol
+        self.ncol = awk_command.ncol_selected
 
         cols_str = np.array2string(cols,separator="\n")
         cols_str = cols_str[1:len(cols_str)-1]
@@ -426,7 +556,7 @@ class DataFrame:
         path_cols = os.path.expanduser('~') + "/.tmp/cols_" + self.id + "_" + awk_command.id + ".txt"
         command = "echo '" + cols_str + "' | sort | xargs -I {} echo \"{}\" >" + path_cols
         os.system(command)
-        variables = "-v header=\"has_header\" -v min_col=\"" + str(min(cols)) + "\" -v max_col=\"" + str(max(cols)) + "\" "
+        variables = "-v has_header=\"has_header_variable\" -v output_header=\"output_header_variable\" -v min_col=\"" + str(min(cols)) + "\" -v max_col=\"" + str(max(cols)) + "\" "
         command = "awk " + variables + """'
         BEGIN {
             FS = \"""" + self.delimiter + """\"
@@ -442,9 +572,20 @@ class DataFrame:
             close(cmd)
         }
         {
-
-            if (FNR == 1){
-                if (header){
+            if (has_header){
+                if (FNR == 1){
+                    if (output_header){
+                        for (i=min_col+1;i<max_col+1;i++){
+                            if (cols[i] == 1){
+                                printf("%s""" + self.delimiter + """", $i)
+                            }
+                        }
+                        if (cols[max_col+1] == 1){
+                            printf("%s", $(max_col+1))
+                        }
+                        printf("%s" RS_new,"")
+                    }
+                }else{
                     for (i=min_col+1;i<max_col+1;i++){
                         if (cols[i] == 1){
                             printf("%s""" + self.delimiter + """", $i)
@@ -454,6 +595,7 @@ class DataFrame:
                         printf("%s", $(max_col+1))
                     }
                     printf("%s" RS_new,"")
+
                 }
             }else{
                 for (i=min_col+1;i<max_col+1;i++){
@@ -465,7 +607,6 @@ class DataFrame:
                     printf("%s", $(max_col+1))
                 }
                 printf("%s" RS_new,"")
-
             }
 
         }
@@ -474,6 +615,14 @@ class DataFrame:
         }
         ' """
         awk_command.command = command
+        if output_header:
+                awk_command.output_header = 1
+        else:
+            awk_command.output_header = 0
+        if has_header:
+                awk_command.has_header = 1
+        else:
+            awk_command.has_header = 0
         awk_command.priority = 2
         awk_command.type = "get_cols"
         self.commands.append(awk_command)
@@ -618,8 +767,9 @@ class DataFrame:
         self.commands.append(awk_command)
         return self
 
-    def select(self,condition_as_string):
-
+    def select(self,condition_as_string,has_header = True,output_header = True):
+        self.nrow_modified_to_unknown_value = True
+        
         condition_as_string = condition_as_string.replace(" in ["," _in_ [")
 
         operators = [">=","<=",">","<","==","!=","~"," ","&&","\|\|","&","\|","!","(",")"]
@@ -680,9 +830,7 @@ class DataFrame:
         bit_operators_and_or_not = ["&","|","!"]
         awk_operators_and_or_not = ["&&","||","!"]
         operators = [">","<",">=","<=","==","!=","~"]
-        names = self.__names_current__()
-        names = self.values(clear = False)
-        names = np.array(names.replace("\n","").split(self.delimiter))
+        names = self.columns
         for i in range(len(names)):
             condition_as_string = condition_as_string.replace(names[i],"$" + str(i+1))
 
@@ -783,9 +931,7 @@ class DataFrame:
         bit_operators_and_or_not = ["&","|","!"]
         awk_operators_and_or_not = ["&&","||","!"]
         operators = [">","<",">=","<=","==","!=","~"]
-        names = self.__names_current__()
-        names = self.values(clear = False)
-        names = np.array(names.replace("\n","").split(self.delimiter))
+        names = self.columns
         for i in range(len(names)):
             equation = equation.replace(names[i],"$" + str(i+1) + "")
 
@@ -839,14 +985,31 @@ class DataFrame:
         self.clear_commands()
         return __ddf__
 
-    def to_csv(self,path_output):
-        command = "xargs -I {} echo {}>" + path_output
-        awk_command = Awk_command()
-        awk_command.command = command
-        awk_command.priority = 999999999999
-        awk_command.type = "to_csv"
-        awk_command.persistance_in_time = "instance"
-        self.commands.append(awk_command)
-        self.values()
-        self.path = path_output
-        self.clear_all_commands()
+    def to_csv(self,path_output,append=False,clear_all_commands = True,set_as_new_path = True):
+        if not append:
+            command = "xargs -I {} echo {}>" + path_output
+            awk_command = Awk_command()
+            awk_command.command = command
+            awk_command.priority = 999999999999
+            awk_command.type = "to_csv"
+            awk_command.persistance_in_time = "instance"
+            self.commands.append(awk_command)
+            self.values()
+            
+            if set_as_new_path:
+                self.path = path_output
+            if clear_all_commands:                
+                self.clear_all_commands()
+        else:
+            command = "xargs -I {} echo {}>>" + path_output
+            awk_command = Awk_command()
+            awk_command.command = command
+            awk_command.priority = 999999999999
+            awk_command.type = "to_csv"
+            awk_command.persistance_in_time = "instance"
+            self.commands.append(awk_command)
+            self.values()
+            if set_as_new_path:
+                self.path = path_output               
+            if clear_all_commands:                
+                self.clear_all_commands()
